import pg from 'pg';
import { VectorStoreBaseParams, BaseVectorStore, IsomorphicDB, VectorStoreQuery, VectorStoreQueryResult } from '@llamaindex/core/vector-store';
import { VercelPool } from '@vercel/postgres';
import { Sql } from 'postgres';
import { BaseEmbedding } from '@llamaindex/core/embeddings';
import { BaseNode, Metadata, StoredValue } from '@llamaindex/core/schema';
import * as _llamaindex_core_storage_doc_store from '@llamaindex/core/storage/doc-store';
import { KVDocumentStore } from '@llamaindex/core/storage/doc-store';
import { BaseKVStore } from '@llamaindex/core/storage/kv-store';
import { KVIndexStore } from '@llamaindex/core/storage/index-store';

declare const PGVECTOR_SCHEMA = "public";
declare const PGVECTOR_TABLE = "llamaindex_embedding";
declare const DEFAULT_DIMENSIONS = 1536;
type PGVectorStoreBaseConfig = {
    schemaName?: string | undefined;
    tableName?: string | undefined;
    dimensions?: number | undefined;
    embedModel?: BaseEmbedding | undefined;
    performSetup?: boolean | undefined;
};
type PGVectorStoreConfig = VectorStoreBaseParams & PGVectorStoreBaseConfig & ({
    /**
     * Client configuration options for the pg client.
     *
     * {@link https://node-postgres.com/apis/client#new-client PostgresSQL Client API}
     */
    clientConfig: pg.ClientConfig;
} | {
    /**
     * A pg client or pool client instance.
     * If provided, make sure it is not connected to the database yet, or it will throw an error.
     */
    shouldConnect?: boolean | undefined;
    client: pg.Client | pg.PoolClient;
} | {
    /**
     * No need to connect to the database, the client is already connected.
     */
    shouldConnect?: false;
    client: Sql | VercelPool;
});
/**
 * Provides support for writing and querying vector data in Postgres.
 * Note: Can't be used with data created using the Python version of the vector store (https://docs.llamaindex.ai/en/stable/examples/vector_stores/postgres/)
 */
declare class PGVectorStore extends BaseVectorStore {
    storesText: boolean;
    private collection;
    private readonly schemaName;
    private readonly tableName;
    private readonly dimensions;
    private isDBConnected;
    private db;
    private readonly clientConfig;
    private readonly performSetup;
    constructor(config: PGVectorStoreConfig);
    /**
     * Setter for the collection property.
     * Using a collection allows for simple segregation of vector data,
     * e.g. by user, source, or access-level.
     * Leave/set blank to ignore the collection value when querying.
     * @param coll Name for the collection.
     */
    setCollection(coll: string): void;
    /**
     * Getter for the collection property.
     * Using a collection allows for simple segregation of vector data,
     * e.g. by user, source, or access-level.
     * Leave/set blank to ignore the collection value when querying.
     * @returns The currently-set collection value.  Default is empty string.
     */
    getCollection(): string;
    private getDb;
    private checkSchema;
    /**
     * Connects to the database specified in environment vars.
     * This method also checks and creates the vector extension,
     * the destination table and indexes if not found.
     * @returns A connection to the database, or the error encountered while connecting/setting up.
     */
    client(): Promise<IsomorphicDB>;
    /**
     * Delete all vector records for the specified collection.
     * NOTE: Uses the collection property controlled by setCollection/getCollection.
     * @returns The result of the delete query.
     */
    clearCollection(): Promise<any[]>;
    private getDataToInsert;
    /**
     * Adds vector record(s) to the table.
     * NOTE: Uses the collection property controlled by setCollection/getCollection.
     * @param embeddingResults The Nodes to be inserted, optionally including metadata tuples.
     * @returns A list of zero or more id values for the created records.
     */
    add(embeddingResults: BaseNode<Metadata>[]): Promise<string[]>;
    /**
     * Deletes a single record from the database by id.
     * NOTE: Uses the collection property controlled by setCollection/getCollection.
     * @param refDocId Unique identifier for the record to delete.
     * @param deleteKwargs Required by VectorStore interface.  Currently ignored.
     * @returns Promise that resolves if the delete query did not throw an error.
     */
    delete(refDocId: string, deleteKwargs?: object): Promise<void>;
    private toPostgresCondition;
    private toPostgresOperator;
    private buildFilterClause;
    /**
     * Query the vector store for the closest matching data to the query embeddings
     * @param query The VectorStoreQuery to be used
     * @param options Required by VectorStore interface.  Currently ignored.
     * @returns Zero or more Document instances with data from the vector store.
     */
    query(query: VectorStoreQuery, options?: object): Promise<VectorStoreQueryResult>;
    /**
     * Required by VectorStore interface.  Currently ignored.
     * @param persistPath
     * @returns Resolved Promise.
     */
    persist(persistPath: string): Promise<void>;
    private escapeLikeString;
}

type DataType = Record<string, Record<string, StoredValue>>;
type PostgresKVStoreBaseConfig = {
    schemaName?: string | undefined;
    tableName?: string | undefined;
};
type PostgresKVStoreClientConfig = {
    /**
     * Client configuration options for the pg client.
     *
     * {@link https://node-postgres.com/apis/client#new-client PostgresSQL Client API}
     */
    clientConfig?: pg.ClientConfig | undefined;
} | {
    /**
     * A pg client or pool client instance.
     * If provided, make sure it is not connected to the database yet, or it will throw an error.
     */
    shouldConnect?: boolean | undefined;
    client?: pg.Client | pg.PoolClient;
};
type PostgresKVStoreConfig = PostgresKVStoreBaseConfig & PostgresKVStoreClientConfig;
declare class PostgresKVStore extends BaseKVStore {
    private schemaName;
    private tableName;
    private isDBConnected;
    private clientConfig;
    private db?;
    constructor(config?: PostgresKVStoreConfig);
    private getDb;
    private checkSchema;
    client(): Promise<pg.ClientBase>;
    put(key: string, val: StoredValue, collection?: string): Promise<void>;
    get(key: string, collection?: string): Promise<any>;
    getAll(collection?: string): Promise<DataType>;
    delete(key: string, collection?: string): Promise<boolean>;
}

type PostgresDocumentStoreConfig = PostgresKVStoreConfig & {
    namespace?: string;
};
declare class PostgresDocumentStore extends KVDocumentStore {
    serializer: _llamaindex_core_storage_doc_store.Serializer<Record<string, unknown>>;
    constructor(config?: PostgresDocumentStoreConfig);
}

type PostgresIndexStoreConfig = PostgresKVStoreConfig & {
    namespace?: string;
};
declare class PostgresIndexStore extends KVIndexStore {
    constructor(config?: PostgresIndexStoreConfig);
}

export { DEFAULT_DIMENSIONS, type DataType, PGVECTOR_SCHEMA, PGVECTOR_TABLE, PGVectorStore, type PGVectorStoreConfig, PostgresDocumentStore, type PostgresDocumentStoreConfig, PostgresIndexStore, type PostgresIndexStoreConfig, PostgresKVStore, type PostgresKVStoreBaseConfig, type PostgresKVStoreClientConfig, type PostgresKVStoreConfig };
